@@ -1,7 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003-2004 Imendio AB
- * Copyright (C) 2003-2004 Sjoerd Simons <sjoerd@luon.net>
+ * Copyright (C) 2003-2005 Imendio AB
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
@@ -77,6 +76,7 @@ struct _LmConnection {
 	guint         io_watch_err;
 	guint         io_watch_hup;
 	guint         fd;
+    guint         io_watch_connect;
 	
 	guint         open_id;
 	LmCallback   *open_cb;
@@ -183,7 +183,8 @@ static GSourceFuncs incoming_funcs = {
 static void
 connection_free (LmConnection *connection)
 {
-	int i;
+	int        i;
+	LmMessage *m;
 
 	g_free (connection->server);
 	g_free (connection->jid);
@@ -203,11 +204,19 @@ connection_free (LmConnection *connection)
 	}
 
 	g_hash_table_destroy (connection->id_handlers);
-
-	if (lm_connection_is_open (connection)) {
+	if (connection->state >= LM_CONNECTION_STATE_OPENING) {
 		connection_do_close (connection);
 	}
 
+	while ((m = g_queue_pop_head (connection->incoming_messages)) != NULL) {
+		lm_message_unref (m);
+	}
+
+	if (connection->ssl) {
+		lm_ssl_unref (connection->ssl);
+	}
+
+	g_queue_free (connection->incoming_messages);
 	g_free (connection);
 }
 
@@ -284,6 +293,12 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 	gchar        *ch;
 
 	connection = connect_data->connection;
+	
+	if (connection->io_watch_connect != 0) {
+		g_source_destroy (g_main_context_find_source_by_id(connection->context,
+								   connection->io_watch_connect));
+		connection->io_watch_connect = 0;
+	}
 
 	/* Need some way to report error/success */
 	if (connection->cancel_open) {
@@ -313,9 +328,8 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 				    NULL)) {
 			shutdown (connection->fd, SHUT_RDWR);
 			close (connection->fd);
+		
 			connection_do_close (connection);
-			connection->fd = -1;
-			g_io_channel_unref(connection->io_channel);
 			return FALSE;
 		}
 	}
@@ -384,17 +398,34 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 void 
 _lm_connection_failed_with_error (LmConnectData *connect_data, int error) 
 {
+	LmConnection *connection;
+	
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
 	       "Connection failed: %s (error %d)\n",
 	       strerror (error), error);
 	
+	connection = connect_data->connection;
+	
 	connect_data->current_addr = connect_data->current_addr->ai_next;
 	
+	if (connection->io_watch_connect != 0) {
+		g_source_destroy (g_main_context_find_source_by_id(connection->context,
+								   connection->io_watch_connect));
+		connection->io_watch_connect = 0;
+	}
+
 	if (connect_data->io_channel != NULL) {
 		g_io_channel_unref (connect_data->io_channel);
 	}
 	
 	if (connect_data->current_addr == NULL) {
+		if (connection->open_cb && connection->open_cb->func) {
+			LmCallback *cb = connection->open_cb;
+			
+			(* ((LmResultFunction) cb->func)) (connection, FALSE,
+							   cb->user_data);
+		}
+
 		freeaddrinfo (connect_data->resolved_addrs);
 		g_free (connect_data);
 	} else {
@@ -414,13 +445,11 @@ connection_connect_cb (GIOChannel   *source,
 		       GIOCondition  condition,
 		       gpointer      data) 
 {
-	LmConnection  *connection;
 	LmConnectData *connect_data;
 	int            error;
 	int            len  = sizeof(error);
 
 	connect_data = (LmConnectData *) data;
-	connection   = connect_data->connection;
 	
 	if (condition == G_IO_ERR) {
 		/* get the real error from the socket */
@@ -492,12 +521,14 @@ connection_do_connect (LmConnectData *connect_data)
 	
 	connect_data->io_channel = g_io_channel_unix_new (fd);
 	if (connection->proxy) {
+        connection->io_watch_connect =
 		connection_add_watch (connection,
 				      connect_data->io_channel,
 				      G_IO_OUT|G_IO_ERR,
 				      (GIOFunc) _lm_proxy_connect_cb, 
 				      connect_data);
 	} else {
+        connection->io_watch_connect =
 		connection_add_watch (connection,
 				      connect_data->io_channel,
 				      G_IO_OUT|G_IO_ERR,
@@ -666,10 +697,19 @@ connection_do_close (LmConnection *connection)
 		g_source_destroy (g_main_context_find_source_by_id (
 			connection->context, connection->io_watch_hup));
 
+		if (connection->io_watch_connect != 0) {
+			g_source_destroy (g_main_context_find_source_by_id(connection->context,
+									   connection->io_watch_connect));
+			connection->io_watch_connect = 0;
+		}
+
 		g_io_channel_unref (connection->io_channel);
 		connection->io_channel = NULL;
+
+		connection->fd = -1;
 	}
 
+	
 	g_source_destroy (connection->incoming_source);
 	g_source_unref (connection->incoming_source);
 
@@ -691,7 +731,7 @@ connection_in_event (GIOChannel   *source,
 		     LmConnection *connection)
 {
 	gchar     buf[IN_BUFFER_SIZE];
-	gsize      bytes_read;
+	gsize     bytes_read;
 	GIOStatus status;
        
 	if (!connection->io_channel) {
@@ -708,7 +748,7 @@ connection_in_event (GIOChannel   *source,
 						  NULL);
 	}
 
-	if (status != G_IO_STATUS_NORMAL) {
+	if (status != G_IO_STATUS_NORMAL || bytes_read < 0) {
 		gint reason;
 		
 		switch (status) {
